#include "hoops_license.h"
#include "HDB.h"
#include "HBaseModel.h"
#include "HBaseView.h"
#include "HIOManager.h"
#include "PointCloudAPI.h"
#include <thread>
#include <vector>


struct ExitCode {
	enum {
		Success = 0,
		Bad_Command_Line_Arguments,
		Could_Not_Load_File,
		OOC_Failure,
	};
};


static ooc::Env load_ooc (HBaseView & view, char const * file_name)
{
	wchar_t w_file_name[4096];
	std::copy_n(file_name, strlen(file_name) + 1, w_file_name);

	HC_KEY destination = view.GetModelKey();

	HInputHandlerOptions options;
	options.m_pHBaseView = &view;
	options.m_bOOCAutoUpdate = false; // Manually control rendering of OOC data for demo code.

	ooc::io::IOResult result = ooc::io::FileInputByKey(w_file_name, destination, options);
	if (result != ooc::io::IOResult_Success) {
		return ooc::Env(); // Return an invalid Env.
	}

	HC_KEY ooc_root = HC_Create_Segment_Key_By_Key(destination, "root");
	ooc::Env env;
	ooc::GetEnv(ooc_root, env);
	return env;
}


static void update (HBaseView & view, ooc::Env env)
{
	view.Update(); // Trigger OOC nodes to start populating.

	while (ooc::PollNodesLoadingOrHaveBeenLoaded(env)) {
		std::this_thread::sleep_for(std::chrono::milliseconds(50));
		view.Update(); // Render any newly populated OOC nodes.
	}
}


static std::vector<ooc::NodeHandle> get_nodes_to_delete (ooc::Env env)
{
	std::vector<ooc::NodeHandle> node_handles;

	HC_KEY ooc_root = ooc::GetRoot(env);
	ooc::NodeHandle root_handle;
	ooc::GetNodeHandle(env, ooc_root, root_handle);
	node_handles.push_back(root_handle);

	HC_Open_Segment_By_Key(ooc_root);{
		HC_Begin_Contents_Search("...", "segment");{
			HC_KEY seg;
			while (HC_Find_Contents(nullptr, &seg)) {
				ooc::NodeHandle node_handle;
				ooc::GetNodeHandle(env, seg, node_handle);
				node_handles.push_back(node_handle);
			}
		}HC_End_Contents_Search();
	}HC_Close_Segment();

	return node_handles;
}


static bool decimate_node (float fraction, ooc::delta::SyncToken const & sync_token, ooc::NodeHandle node_handle)
{
	using ooc::delta::ModifyResult;

	ooc::Env env = sync_token.GetEnv();
	int initial_point_count = ooc::query::PointCount(env, node_handle);
	int kill_point_count = static_cast<int>(fraction * initial_point_count);

	std::vector<int> point_indices(kill_point_count);
	for (int i = 0; i < kill_point_count; ++i) {
		point_indices[i] = i;
	}

	ModifyResult modify_res = ooc::delta::DeleteSpecificPoints(
		sync_token, node_handle,
		point_indices.data(), point_indices.size());

	return modify_res == ooc::delta::ModifyResult_Success;
}


static bool decimate_nodes (float fraction, ooc::Env env, std::vector<ooc::NodeHandle> const & node_handles)
{
	
	using ooc::delta::SyncResult;
	using ooc::delta::SyncToken;
	using ooc::delta::SynchronizeWith;

	bool success = true;

	SyncResult sync_res = SynchronizeWith(env, [&](SyncToken const & sync_token) {
		for (size_t i = 0; i < node_handles.size(); ++i) {
			ooc::NodeHandle node_handle = node_handles[i];
			if (!decimate_node(fraction, sync_token, node_handle)) {
				success = false;
				return;
			}
		}
	});

	if (sync_res != ooc::delta::SyncResult_Success) {
		success = false;
	}

	return success;
}


int main (int argc, char ** args)
{
	if (argc != 2) {
		return ExitCode::Bad_Command_Line_Arguments;
	}

	HC_Define_System_Options("license = `" HOOPS_LICENSE "`");
	HC_Define_System_Options("multi-threading = full"); // Multi-threading needs to be enabled for asynchronous (default) OOC use.

	HDB hdb;
	hdb.Init();
	HBaseModel model;
	model.Init();
	HBaseView view(&model);
	view.Init();

	char const * file_name = args[1];
	ooc::Env env = load_ooc(view, file_name);
	if (env == ooc::Env()) {
		return ExitCode::Could_Not_Load_File;
	}

	float kill_points_fraction = 0.85f; // Delete this fraction of points in all OOC nodes.
	if (!decimate_nodes(kill_points_fraction, env, get_nodes_to_delete(env))) {
		return ExitCode::OOC_Failure;
	}

	update(view, env);

	HC_Define_Alias("?Picture", view.GetDriverPath());
	HC_Pause();

	return ExitCode::Success;
}





